@@ -63,6 +63,22 @@ STATUS_DEFAULT = {
 
 class MainWindow(object):
     def __init__(self, root):
+        '''
+        -----------------------------------------------------
+        | main button toolbar                               |
+        -----------------------------------------------------
+        |       < ma | in content area >                    |
+        |            |                                      |
+        |  left      |              right                   |
+        |  control   |              details frame           |
+        |  tree      |              / output viewer         |
+        |  area      |                                      |
+        -----------------------------------------------------
+        |     status bar area                               |
+        -----------------------------------------------------
+
+        '''
+
         self._project = None
         self.executor = None
 
@@ -77,6 +93,38 @@ class MainWindow(object):
         self.root.protocol("WM_DELETE_WINDOW", self.on_quit)
         # Catch the "quit" event.
         self.root.createcommand('exit', self.on_quit)
+
+        # Setup the menu
+        self._setup_menubar()
+
+        # Set up the main content for the window.
+        self._setup_button_toolbar()
+        self._setup_main_content()
+        self._setup_status_bar()
+
+        # Now configure the weights for the root frame
+        self.root.columnconfigure(0, weight=1)
+        self.root.rowconfigure(0, weight=0)
+        self.root.rowconfigure(1, weight=1)
+        self.root.rowconfigure(2, weight=0)
+
+        # Set up listeners for runner events.
+        Executor.bind('test_status_update', self.on_executorStatusUpdate)
+        Executor.bind('test_start', self.on_executorTestStart)
+        Executor.bind('test_end', self.on_executorTestEnd)
+        Executor.bind('suite_end', self.on_executorSuiteEnd)
+        Executor.bind('suite_error', self.on_executorSuiteError)
+
+        # Now that we've laid out the grid, hide the error and output text
+        # until we actually have an error/output to display
+        self._hide_test_output()
+        self._hide_test_errors()
+
+    ######################################################
+    # Internal GUI layout methods.
+    ######################################################
+
+    def _setup_menubar(self):
         # Menubar
         self.menubar = Menu(self.root)
 
@@ -107,6 +155,12 @@ class MainWindow(object):
         # last step - configure the menubar
         self.root['menu'] = self.menubar
 
+    def _setup_button_toolbar(self):
+        '''
+        The button toolbar runs as a horizontal area at the top of the GUI.
+        It is a persistent GUI component
+        '''
+
         # Main toolbar
         self.toolbar = Frame(self.root)
         self.toolbar.grid(column=0, row=0, sticky=(W, E))
@@ -118,21 +172,51 @@ class MainWindow(object):
         self.run_all_button = Button(self.toolbar, text='Run all', command=self.on_run_all)
         self.run_all_button.grid(column=1, row=0)
 
-        self.run_selected_button = Button(self.toolbar, text='Run selected', command=self.on_run_selected)
+        self.run_selected_button = Button(self.toolbar, text='Run selected',
+            command=self.on_run_selected, state=DISABLED)
         self.run_selected_button.grid(column=2, row=0)
 
         self.rerun_button = Button(self.toolbar, text='Re-run', command=self.on_rerun, state=DISABLED)
         self.rerun_button.grid(column=3, row=0)
 
+        self.toolbar.columnconfigure(0, weight=0)
+        self.toolbar.rowconfigure(0, weight=1)
+
+    def _setup_main_content(self):
+        '''
+        Sets up the main content area. It is a persistent GUI component
+        '''
+
         # Main content area
         self.content = PanedWindow(self.root, orient=HORIZONTAL)
         self.content.grid(column=0, row=1, sticky=(N, S, E, W))
 
+        # Create the tree/control area on the left frame
+        self._setup_left_frame()
+        self._setup_all_tests_tree()
+        self._setup_problem_tests_tree()
+
+        # Create the output/viewer area on the right frame
+        self._setup_right_frame()
+
+        # Set up weights for the left frame's content
+        self.content.columnconfigure(0, weight=1)
+        self.content.rowconfigure(0, weight=1)
+
+        self.content.pane(0, weight=1)
+        self.content.pane(1, weight=2)
+
+    def _setup_left_frame(self):
+        '''
+        The left frame mostly consists of the tree widget
+        '''
+
         # The left-hand side frame on the main content area
         # The tabs for the two trees
         self.tree_notebook = Notebook(self.content, padding=(0, 5, 0, 5))
         self.content.add(self.tree_notebook)
 
+    def _setup_all_tests_tree(self):
         # The tree for all tests
         self.all_tests_tree_frame = Frame(self.content)
         self.all_tests_tree_frame.grid(column=0, row=0, sticky=(N, S, E, W))
@@ -164,32 +248,48 @@ class MainWindow(object):
         self.all_tests_tree.config(yscrollcommand=self.all_tests_tree_scrollbar.set)
         self.all_tests_tree_scrollbar.config(command=self.all_tests_tree.yview)
 
-        # The tree for problem tests
-        self.problems_tree_frame = Frame(self.content)
-        self.problems_tree_frame.grid(column=0, row=0, sticky=(N, S, E, W))
-        self.tree_notebook.add(self.problems_tree_frame, text='Problems')
+        # Setup weights for the "All Tests" tree
+        self.all_tests_tree_frame.columnconfigure(0, weight=1)
+        self.all_tests_tree_frame.columnconfigure(1, weight=0)
+        self.all_tests_tree_frame.rowconfigure(0, weight=1)
 
-        self.problems_tree = Treeview(self.problems_tree_frame)
-        self.problems_tree.grid(column=0, row=0, sticky=(N, S, E, W))
+    def _setup_problem_tests_tree(self):
+        # The tree for problem tests
+        self.problem_tests_tree_frame = Frame(self.content)
+        self.problem_tests_tree_frame.grid(column=0, row=0, sticky=(N, S, E, W))
+        self.tree_notebook.add(self.problem_tests_tree_frame, text='Problems')
+
+        self.problem_tests_tree = Treeview(self.problem_tests_tree_frame)
+        self.problem_tests_tree.grid(column=0, row=0, sticky=(N, S, E, W))
 
         # Set up the tag colors for tree nodes.
         for status, config in STATUS.items():
-            self.problems_tree.tag_configure(config['tag'], foreground=config['color'])
-        self.problems_tree.tag_configure('inactive', foreground='lightgray')
+            self.problem_tests_tree.tag_configure(config['tag'], foreground=config['color'])
+        self.problem_tests_tree.tag_configure('inactive', foreground='lightgray')
 
         # Problem tree only deals with selection, not clicks.
-        self.problems_tree.tag_bind('TestModule', '<<TreeviewSelect>>', self.on_testModuleSelected)
-        self.problems_tree.tag_bind('TestCase', '<<TreeviewSelect>>', self.on_testCaseSelected)
-        self.problems_tree.tag_bind('TestMethod', '<<TreeviewSelect>>', self.on_testMethodSelected)
+        self.problem_tests_tree.tag_bind('TestModule', '<<TreeviewSelect>>', self.on_testModuleSelected)
+        self.problem_tests_tree.tag_bind('TestCase', '<<TreeviewSelect>>', self.on_testCaseSelected)
+        self.problem_tests_tree.tag_bind('TestMethod', '<<TreeviewSelect>>', self.on_testMethodSelected)
 
         # The tree's vertical scrollbar
-        self.problems_tree_scrollbar = Scrollbar(self.problems_tree_frame, orient=VERTICAL)
-        self.problems_tree_scrollbar.grid(column=1, row=0, sticky=(N, S))
+        self.problem_tests_tree_scrollbar = Scrollbar(self.problem_tests_tree_frame, orient=VERTICAL)
+        self.problem_tests_tree_scrollbar.grid(column=1, row=0, sticky=(N, S))
 
         # Tie the scrollbar to the text views, and the text views
         # to each other.
-        self.problems_tree.config(yscrollcommand=self.problems_tree_scrollbar.set)
-        self.problems_tree_scrollbar.config(command=self.all_tests_tree.yview)
+        self.problem_tests_tree.config(yscrollcommand=self.problem_tests_tree_scrollbar.set)
+        self.problem_tests_tree_scrollbar.config(command=self.all_tests_tree.yview)
+
+        # Setup weights for the problems tree
+        self.problem_tests_tree_frame.columnconfigure(0, weight=1)
+        self.problem_tests_tree_frame.columnconfigure(1, weight=0)
+        self.problem_tests_tree_frame.rowconfigure(0, weight=1)
+
+    def _setup_right_frame(self):
+        '''
+        The right frame is basically the "output viewer" space
+        '''
 
         # The right-hand side frame on the main content area
         self.details_frame = Frame(self.content)
@@ -259,6 +359,19 @@ class MainWindow(object):
         self.error.config(yscrollcommand=self.error_scrollbar.set)
         self.error_scrollbar.config(command=self.error.yview)
 
+        # Set up GUI weights for the details frame
+        self.details_frame.columnconfigure(0, weight=0)
+        self.details_frame.columnconfigure(1, weight=1)
+        self.details_frame.columnconfigure(2, weight=0)
+        self.details_frame.columnconfigure(3, weight=0)
+        self.details_frame.columnconfigure(4, weight=0)
+        self.details_frame.rowconfigure(0, weight=0)
+        self.details_frame.rowconfigure(1, weight=0)
+        self.details_frame.rowconfigure(2, weight=1)
+        self.details_frame.rowconfigure(3, weight=5)
+        self.details_frame.rowconfigure(4, weight=10)
+
+    def _setup_status_bar(self):
         # Status bar
         self.statusbar = Frame(self.root)
         self.statusbar.grid(column=0, row=2, sticky=(W, E))
@@ -280,61 +393,29 @@ class MainWindow(object):
         self.progress = Progressbar(self.statusbar, orient=HORIZONTAL, length=200, mode='determinate', maximum=100, variable=self.progress_value)
         self.progress.grid(column=2, row=0, sticky=(W, E))
 
-        # Set up listeners for runner events.
-        Executor.bind('test_status_update', self.on_executorStatusUpdate)
-        Executor.bind('test_start', self.on_executorTestStart)
-        Executor.bind('test_end', self.on_executorTestEnd)
-        Executor.bind('suite_end', self.on_executorSuiteEnd)
-        Executor.bind('suite_error', self.on_executorSuiteError)
-
         # Main window resize handle
         self.grip = Sizegrip(self.statusbar)
         self.grip.grid(column=3, row=0, sticky=(S, E))
 
-        # Now configure the weights for the frame grids
-        self.root.columnconfigure(0, weight=1)
-        self.root.rowconfigure(0, weight=0)
-        self.root.rowconfigure(1, weight=1)
-        self.root.rowconfigure(2, weight=0)
-
-        self.toolbar.columnconfigure(0, weight=0)
-        self.toolbar.rowconfigure(0, weight=1)
-
-        self.content.columnconfigure(0, weight=1)
-        self.content.rowconfigure(0, weight=1)
-
-        self.content.pane(0, weight=1)
-        self.content.pane(1, weight=2)
-
+        # Set up weights for status bar frame
         self.statusbar.columnconfigure(0, weight=1)
         self.statusbar.columnconfigure(1, weight=0)
         self.statusbar.columnconfigure(2, weight=0)
         self.statusbar.columnconfigure(3, weight=0)
         self.statusbar.rowconfigure(0, weight=1)
 
-        self.all_tests_tree_frame.columnconfigure(0, weight=1)
-        self.all_tests_tree_frame.columnconfigure(1, weight=0)
-        self.all_tests_tree_frame.rowconfigure(0, weight=1)
+    ######################################################
+    # Utility methods for inspecting current GUI state
+    ######################################################
 
-        self.problems_tree_frame.columnconfigure(0, weight=1)
-        self.problems_tree_frame.columnconfigure(1, weight=0)
-        self.problems_tree_frame.rowconfigure(0, weight=1)
-
-        self.details_frame.columnconfigure(0, weight=0)
-        self.details_frame.columnconfigure(1, weight=1)
-        self.details_frame.columnconfigure(2, weight=0)
-        self.details_frame.columnconfigure(3, weight=0)
-        self.details_frame.columnconfigure(4, weight=0)
-        self.details_frame.rowconfigure(0, weight=0)
-        self.details_frame.rowconfigure(1, weight=0)
-        self.details_frame.rowconfigure(2, weight=1)
-        self.details_frame.rowconfigure(3, weight=5)
-        self.details_frame.rowconfigure(4, weight=10)
-
-        # Now that we've laid out the grid, hide the error and output text
-        # until we actually have an error/output to display
-        self._hide_test_output()
-        self._hide_test_errors()
+    @property
+    def current_test_tree(self):
+        "Check the tree notebook to return the currently selected tree."
+        current_tree_id = self.tree_notebook.select()
+        if current_tree_id == self.problem_tests_tree_frame._w:
+            return self.problem_tests_tree
+        else:
+            return self.all_tests_tree
 
     ######################################################
     # Handlers for setting a new project
@@ -453,6 +534,9 @@ class MainWindow(object):
         self._hide_test_output()
         self._hide_test_errors()
 
+        # update "run selected" button enabled state
+        self.set_selected_button_state()
+
     def on_testCaseSelected(self, event):
         "Event handler: a test case has been selected in the tree"
         self.name.set('')
@@ -464,6 +548,9 @@ class MainWindow(object):
         self._hide_test_output()
         self._hide_test_errors()
 
+        # update "run selected" button enabled state
+        self.set_selected_button_state()
+
     def on_testMethodSelected(self, event):
         "Event handler: a test case has been selected in the tree"
         if len(event.widget.selection()) == 1:
@@ -515,6 +602,9 @@ class MainWindow(object):
             self._hide_test_output()
             self._hide_test_errors()
 
+        # update "run selected" button enabled state
+        self.set_selected_button_state()
+
     def on_nodeAdded(self, node):
         "Event handler: a new node has been added to the tree"
         self.all_tests_tree.insert(
@@ -548,8 +638,8 @@ class MainWindow(object):
                 path = '.'.join(parts[:pos+1])
                 testModule = parentModule[part]
 
-                if not self.problems_tree.exists(path):
-                    self.problems_tree.insert(
+                if not self.problem_tests_tree.exists(path):
+                    self.problem_tests_tree.insert(
                         parentModule.path, 'end', testModule.path,
                         text=testModule.name,
                         tags=[testModule.__class__.__name__, 'active'],
@@ -558,19 +648,19 @@ class MainWindow(object):
 
                 parentModule = testModule
 
-            self.problems_tree.item(node.path, tags=['TestMethod', STATUS[node.status]['tag']])
+            self.problem_tests_tree.item(node.path, tags=['TestMethod', STATUS[node.status]['tag']])
         else:
             # Test passed; if it's on the problem tree, remove it.
-            if self.problems_tree.exists(node.path):
-                self.problems_tree.delete(node.path)
+            if self.problem_tests_tree.exists(node.path):
+                self.problem_tests_tree.delete(node.path)
 
                 # Check all parents of this node. Recursively remove
                 # any parent has no children as a result of this deletion.
                 has_children = False
                 node = node.parent
                 while node.path and not has_children:
-                    if not self.problems_tree.get_children(node.path):
-                        self.problems_tree.delete(node.path)
+                    if not self.problem_tests_tree.get_children(node.path):
+                        self.problem_tests_tree.delete(node.path)
                     else:
                         has_children = True
                     node = node.parent
@@ -588,13 +678,7 @@ class MainWindow(object):
 
     def on_run_selected(self, event=None):
         "Event handler: The 'run selected' button has been pressed"
-        # Check the tree notebook to see which tree
-        # is currently selected.
-        current_tree_id = self.tree_notebook.select()
-        if current_tree_id == self.problems_tree_frame._w:
-            current_tree = self.problems_tree
-        else:
-            current_tree = self.all_tests_tree
+        current_tree = self.current_test_tree
 
         # If a node is selected, it needs to be made active
         for path in current_tree.selection():
@@ -651,12 +735,7 @@ class MainWindow(object):
 
         # If the test that just fininshed is the one (and only one)
         # selected on the tree, update the display.
-        current_tree_id = self.tree_notebook.select()
-        if current_tree_id == self.problems_tree_frame._w:
-            current_tree = self.problems_tree
-        else:
-            current_tree = self.all_tests_tree
-
+        current_tree = self.current_test_tree
         if len(current_tree.selection()) == 1:
             # One test selected.
             if current_tree.selection()[0] == test_path:
@@ -675,11 +754,15 @@ class MainWindow(object):
             self._hide_test_output()
             self._hide_test_errors()
 
-    def on_executorSuiteEnd(self, event):
+    def on_executorSuiteEnd(self, event, error=None):
         "The test suite finished running."
         # Display the final results
         self.run_status.set('Finished.')
-        if sum(self.executor.result_count.get(state, 0) for state in TestMethod.FAILING_STATES):
+
+        if error:
+            TestErrorsDialog(self.root, error)
+
+        if self.executor.any_failed:
             dialog = tkMessageBox.showerror
         else:
             dialog = tkMessageBox.showinfo
@@ -700,10 +783,7 @@ class MainWindow(object):
             })
 
         # Reset the buttons
-        self.stop_button.configure(state=DISABLED)
-        self.run_all_button.configure(state=NORMAL)
-        self.run_selected_button.configure(state=NORMAL)
-        self.rerun_button.configure(state=NORMAL)
+        self.reset_button_states_on_end()
 
         # Drop the reference to the executor
         self.executor = None
@@ -715,14 +795,30 @@ class MainWindow(object):
         FailedTestDialog(self.root, error)
 
         # Reset the buttons
-        self.stop_button.configure(state=DISABLED)
-        self.run_all_button.configure(state=NORMAL)
-        self.run_selected_button.configure(state=NORMAL)
-        self.rerun_button.configure(state=NORMAL)
+        self.reset_button_states_on_end()
 
         # Drop the reference to the executor
         self.executor = None
 
+    def reset_button_states_on_end(self):
+        "A test run has ended and we should enable or disable buttons as appropriate."
+        self.stop_button.configure(state=DISABLED)
+        self.run_all_button.configure(state=NORMAL)
+        self.set_selected_button_state()
+        if self.executor.any_failed:
+            self.rerun_button.configure(state=NORMAL)
+        else:
+            self.rerun_button.configure(state=DISABLED)
+
+    def set_selected_button_state(self):
+        if self.executor and self.executor.is_running:
+            self.run_selected_button.configure(state=DISABLED)
+        elif self.current_test_tree.selection():
+            self.run_selected_button.configure(state=NORMAL)
+        else:
+            self.run_selected_button.configure(state=DISABLED)
+
+
     ######################################################
     # GUI utility methods
     ######################################################
@@ -765,10 +861,7 @@ class MainWindow(object):
 
             self.run_status.set('Stopped.')
 
-            self.stop_button.configure(state=DISABLED)
-            self.run_all_button.configure(state=NORMAL)
-            self.run_selected_button.configure(state=NORMAL)
-            self.rerun_button.configure(state=NORMAL)
+            self.reset_button_states_on_end()
 
     def _hide_test_output(self):
         "Hide the test output panel on the test results page"
@@ -850,8 +943,9 @@ class StackTraceDialog(Toplevel):
 
         self.description.insert('1.0', trace)
 
-        self.cancel_button = Button(self.frame, text=cancel_text, command=self.cancel)
-        self.cancel_button.grid(column=0, row=2, padx=5, pady=5, sticky=(E,))
+        if cancel_text is not None:
+            self.cancel_button = Button(self.frame, text=cancel_text, command=self.cancel)
+            self.cancel_button.grid(column=0, row=2, padx=5, pady=5, sticky=(E,))
 
         self.ok_button = Button(self.frame, text=button_text, command=self.ok, default=ACTIVE)
         self.ok_button.grid(column=1, row=2, padx=5, pady=5, sticky=(E,))
@@ -927,6 +1021,30 @@ class FailedTestDialog(StackTraceDialog):
         self.parent.quit()
 
 
+class TestErrorsDialog(StackTraceDialog):
+    def __init__(self, parent, trace):
+        '''Show a dialog with a scrollable list of errors.
+
+        Arguments:
+
+            parent -- a parent window (the application window)
+            error -- the error content to display.
+        '''
+        StackTraceDialog.__init__(
+            self,
+            parent,
+            'Errors during test suite',
+            ('The following errors were generated while running the test suite:'),
+            trace,
+            button_text='OK',
+            cancel_text=None,
+        )
+
+    def cancel(self, event=None):
+        StackTraceDialog.cancel(self, event=event)
+        self.parent.quit()
+
+
 class TestLoadErrorDialog(StackTraceDialog):
     def __init__(self, parent, trace):
         '''Show a dialog with a scrollable stack trace.
